from ..i18n import _
from ..error import DictConsistencyError


#mode order is important
modes_level = ('basic', 'normal', 'expert')


class Mode(object):
    def __init__(self, name, level):
        self.name = name
        self.level = level
    def __gt__(self, other):
        return other.level < self.level


def mode_factory():
    mode_obj = {}
    for idx in range(len(modes_level)):
        name = modes_level[idx]
        mode_obj[name] = Mode(name, idx)
    return mode_obj

modes = mode_factory()


class FamilyAnnotator:
    def __init__(self,
                 objectspace,
                 ):
        self.objectspace = objectspace
        self.remove_empty_families()
        self.change_variable_mode()
        self.change_family_mode()
        self.dynamic_families()

    def remove_empty_families(self):  # pylint: disable=C0111,R0201
        if hasattr(self.objectspace.space, 'variables'):
            for family in self.objectspace.space.variables.values():
                if hasattr(family, 'family'):
                    space = family.family
                    removed_families = []
                    for family_name, family in space.items():
                        if not hasattr(family, 'variable') or len(family.variable) == 0:
                            removed_families.append(family_name)
                    for family_name in removed_families:
                        del space[family_name]

    def change_family_mode(self):  # pylint: disable=C0111
        if not hasattr(self.objectspace.space, 'variables'):
            return
        for family in self.objectspace.space.variables.values():
            if hasattr(family, 'family'):
                for family in family.family.values():
                    mode = modes_level[-1]
                    for variable in family.variable.values():
                        if isinstance(variable, self.objectspace.leadership):
                            variable_mode = variable.variable[0].mode
                            variable.variable[0].mode = None
                            variable.mode = variable_mode
                        else:
                            variable_mode = variable.mode
                        if variable_mode is not None and modes[mode] > modes[variable_mode]:
                            mode = variable_mode
                    family.mode = mode

    def dynamic_families(self):  # pylint: disable=C0111
        if not hasattr(self.objectspace.space, 'variables'):
            return
        for family in self.objectspace.space.variables.values():
            if hasattr(family, 'family'):
                for family in family.family.values():
                    if 'dynamic' in vars(family):
                        namespace = self.objectspace.paths.get_variable_namespace(family.dynamic)
                        varpath = self.objectspace.paths.get_variable_path(family.dynamic, namespace)
                        family.dynamic = varpath

    def annotate_variable(self, variable, family_mode, path, is_follower=False):
        # if the variable is mandatory and doesn't have any value
        # then the variable's mode is set to 'basic'
        if not hasattr(variable, 'value') and variable.type == 'boolean':
            new_value = self.objectspace.value(variable.xmlfiles)
            new_value.name = True
            new_value.type = 'boolean'
            variable.value = [new_value]
        if hasattr(variable, 'value') and variable.value:
            has_value = True
            for value in variable.value:
                if value.type == 'calculation':
                    has_value = False
                    has_variable = False
                    if hasattr(value, 'param'):
                        for param in value.param:
                            if param.type == 'variable':
                                has_variable = True
                                break
                    #if not has_variable:
                    #    # if one parameter is a variable, let variable choice if it's mandatory
                    #    variable.mandatory = True
            if has_value:
                # if has value but without any calculation
                variable.mandatory = True
        if variable.mandatory is True and (not hasattr(variable, 'value') or is_follower):
            variable.mode = modes_level[0]
        if variable.mode != None and modes[variable.mode] < modes[family_mode] and (not is_follower or variable.mode != modes_level[0]):
            variable.mode = family_mode
        if variable.hidden is True:
            variable.frozen = True
            if not variable.auto_save is True and 'force_default_on_freeze' not in vars(variable):
                variable.force_default_on_freeze = True

    def change_variable_mode(self):  # pylint: disable=C0111
        if not hasattr(self.objectspace.space, 'variables'):
            return
        for variables in self.objectspace.space.variables.values():
            namespace = variables.name
            if hasattr(variables, 'family'):
                for family in variables.family.values():
                    family_mode = family.mode
                    if hasattr(family, 'variable'):
                        for variable in family.variable.values():

                            if isinstance(variable, self.objectspace.leadership):
                                mode = modes_level[-1]
                                for idx, follower in enumerate(variable.variable):
                                    if follower.auto_save is True:
                                        xmlfiles = self.objectspace.display_xmlfiles(variable.xmlfiles)
                                        raise DictConsistencyError(_(f'leader/followers "{follower.name}" could not be auto_save in {xmlfiles}'), 29)
                                    if follower.auto_freeze is True:
                                        xmlfiles = self.objectspace.display_xmlfiles(variable.xmlfiles)
                                        raise DictConsistencyError(_(f'leader/followers "{follower.name}" could not be auto_freeze in {xmlfiles}'), 30)
                                    is_follower = idx != 0
                                    path = '{}.{}.{}'.format(family.path, variable.name, follower.name)
                                    self.annotate_variable(follower, family_mode, path, is_follower)
                                    # leader's mode is minimum level
                                    if modes[variable.variable[0].mode] > modes[follower.mode]:
                                        follower.mode = variable.variable[0].mode
                                variable.mode = variable.variable[0].mode
                            else:
                                # auto_save's variable is set in 'basic' mode if its mode is 'normal'
                                if variable.auto_save is True and variable.mode != modes_level[-1]:
                                    variable.mode = modes_level[0]
                                # auto_freeze's variable is set in 'basic' mode if its mode is 'normal'
                                if variable.auto_freeze is True and variable.mode != modes_level[-1]:
                                    variable.mode = modes_level[0]
                                path = '{}.{}'.format(family.path, variable.name)
                                self.annotate_variable(variable, family_mode, path)

