from os.path import basename

from ..i18n import _
from ..utils import normalize_family
from ..error import DictConsistencyError
# a CreoleObjSpace's attribute has some annotations
# that shall not be present in the exported (flatened) XML
ERASED_ATTRIBUTES = ('redefine', 'exists', 'fallback', 'optional', 'remove_check', 'namespace',
                     'remove_condition', 'path', 'instance_mode', 'index', 'is_in_leadership',
                     'level', 'remove_fill', 'xmlfiles')


KEY_TYPE = {'variable': 'symlink',
            'SymLinkOption': 'symlink',
            'PortOption': 'port',
            'UnicodeOption': 'string',
            'NetworkOption': 'network',
            'NetmaskOption': 'netmask',
            'URLOption': 'web_address',
            'FilenameOption': 'filename'}


class ServiceAnnotator:
    """Manage service's object
    for example::
        <services>
            <service name="test">
                <service_access service='ntp'>
                    <port protocol='udp' service_accesslist='ntp_udp'>123</port>
                </service_access>
            </service>
        </services>
    """
    def __init__(self, objectspace):
        self.objectspace = objectspace
        self.convert_services()

    def convert_services(self):
        if not hasattr(self.objectspace.space, 'services'):
            return
        if not hasattr(self.objectspace.space.services, 'service'):
            del self.objectspace.space.services
            return
        self.objectspace.space.services.hidden = True
        self.objectspace.space.services.name = 'services'
        self.objectspace.space.services.doc = 'services'
        families = {}
        for idx, service_name in enumerate(self.objectspace.space.services.service.keys()):
            service = self.objectspace.space.services.service[service_name]
            new_service = self.objectspace.service(service.xmlfiles)
            for elttype, values in vars(service).items():
                if not isinstance(values, (dict, list)) or elttype in ERASED_ATTRIBUTES:
                    setattr(new_service, elttype, values)
                    continue
                eltname = elttype + 's'
                path = '.'.join(['services', service_name, eltname])
                family = self.gen_family(eltname,
                                         path,
                                         service.xmlfiles,
                                         )
                if isinstance(values, dict):
                    values = list(values.values())
                family.family = self.make_group_from_elts(service_name,
                                                          elttype,
                                                          values,
                                                          path,
                                                          )
                setattr(new_service, elttype, family)
            new_service.doc = new_service.name
            families[service_name] = new_service
        self.objectspace.space.services.service = families

    def gen_family(self,
                   name,
                   path,
                   xmlfiles
                   ):
        family = self.objectspace.family(xmlfiles)
        family.name = normalize_family(name)
        family.doc = name
        family.mode = None
        self.objectspace.paths.add_family('services',
                                          path,
                                          family,
                                          )
        return family

    def make_group_from_elts(self,
                             service_name,
                             name,
                             elts,
                             path,
                             ):
        """Splits each objects into a group (and `OptionDescription`, in tiramisu terms)
        and build elements and its attributes (the `Options` in tiramisu terms)
        """
        families = []
        new_elts = self._reorder_elts(name,
                                      elts,
                                      )
        for index, elt_info in enumerate(new_elts):
            elt = elt_info['elt']
            elt_name = elt_info['elt_name']

            # try to launch _update_xxxx() function
            update_elt = '_update_' + elt_name
            if hasattr(self, update_elt):
                getattr(self, update_elt)(elt,
                                          index,
                                          path,
                                          service_name,
                                          )

            idx = 0
            while True:
                if hasattr(elt, 'source'):
                    c_name = elt.source
                else:
                    c_name = elt.name
                if idx:
                    c_name += f'_{idx}'
                subpath = '{}.{}'.format(path, c_name)
                if not self.objectspace.paths.family_is_defined(subpath):
                    break
                idx += 1
            family = self.gen_family(c_name,
                                     subpath,
                                     elt.xmlfiles,
                                     )
            family.variable = []
            listname = '{}list'.format(name)
            activate_path = '.'.join([subpath, 'activate'])
            for key in dir(elt):
                if key.startswith('_') or key.endswith('_type') or key in ERASED_ATTRIBUTES:
                    continue
                value = getattr(elt, key)
                if key == listname:
                    self.objectspace.list_conditions.setdefault(listname,
                                                                {}).setdefault(
                                                                    value,
                                                                    []).append(activate_path)
                    continue
                family.variable.append(self._generate_element(elt_name,
                                                              key,
                                                              value,
                                                              elt,
                                                              f'{subpath}.{key}'
                                                              ))
            # FIXME ne devrait pas etre True par défaut
            # devrait etre un calcule
            family.variable.append(self._generate_element(elt_name,
                                                         'activate',
                                                          True,
                                                          elt,
                                                          activate_path,
                                                          ))
            families.append(family)
        return families

    def _generate_element(self,
                          elt_name,
                          key,
                          value,
                          elt,
                          path,
                          ):
        variable = self.objectspace.variable(elt.xmlfiles)
        variable.name = normalize_family(key)
        variable.mode = None
        if key == 'name':
            true_key = elt_name
        else:
            true_key = key
        dtd_key_type = true_key + '_type'
        if key == 'activate':
            type_ = 'boolean'
        elif hasattr(elt, dtd_key_type):
            type_ = KEY_TYPE[getattr(elt, dtd_key_type)]
        elif key in self.objectspace.booleans_attributs:
            type_ = 'boolean'
        else:
            type_ = 'string'
        variable.type = type_
        if type_ == 'symlink':
            variable.opt = self.objectspace.paths.get_variable_path(value,
                                                                    'services',
                                                                    )
#            variable.opt = value
            variable.multi = None
        else:
            variable.doc = key
            val = self.objectspace.value(elt.xmlfiles)
            val.type = type_
            val.name = value
            variable.value = [val]
        self.objectspace.paths.add_variable('services',
                                            path,
                                            'service',
                                            False,
                                            variable,
                                            )
        return variable

    def _reorder_elts(self,
                      name,
                      elts,
                      ):
        """Reorders by index the elts
        """
        new_elts = {}
        # reorder elts by index
        for idx, elt in enumerate(elts):
            new_elts.setdefault(idx, []).append(elt)
        idxes = list(new_elts.keys())
        idxes.sort()
        result_elts = []
        for idx in idxes:
            for elt in new_elts[idx]:
                result_elts.append({'elt_name': name, 'elt': elt})
        return result_elts

    def _update_override(self,
                         file_,
                         index,
                         service_path,
                         service_name,
                         ):
        file_.name = f'/systemd/system/{service_name}.service.d/rougail.conf'
        # retrieve default value from File object
        for attr in ['owner', 'group', 'mode']:
            setattr(file_, attr, getattr(self.objectspace.file, attr))
        if not hasattr(file_, 'source'):
            file_.source = f'{service_name}.service'
        self._update_file(file_,
                          index,
                          service_path,
                          service_name,
                          )

    def _update_file(self,
                     file_,
                     index,
                     service_path,
                     service_name,
                     ):
        if not hasattr(file_, 'file_type') or file_.file_type == "UnicodeOption":
            if not hasattr(file_, 'source'):
                file_.source = basename(file_.name)
        elif not hasattr(file_, 'source'):
            xmlfiles = self.objectspace.display_xmlfiles(file_.xmlfiles)
            raise DictConsistencyError(_(f'attribute "source" is mandatory for the file "{file_.name}" in {xmlfiles}'), 34)

